/*++

Copyright (c) 2004 Alexander A. Telyatnikov (Alter)

Module Name:
    PciDumpc.cpp

Abstract:
    This is simple utility that builds full list of PCI devices and
    assigned rosources (IO/Memory ranges, IRQ, etc)

Author:
    Alexander A. Telyatnikov (Alter)

Environment:
    User mode only

Notes:

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Revision History:

--*/

#include <windows.h>
#include <stdio.h>
#include "..\Driver\PCIDump.h"
#include "SvcManLib.h"
#include "PciIds.h"
#include "Privilege.h"

char*
OnUpdateVendorId(
    ULONG VendorID
    ) 
{
    for(int i=0; v_names[i].str; i++) {
        if(v_names[i].id == VendorID) {
            return v_names[i].str;
        }
    }
    return "";
}

char*
OnUpdateDeviceId(
    ULONG VendorID,
    ULONG DeviceID
    ) 
{
    ULONG z = (VendorID << 16) | DeviceID;
	
    for(int i=0; d_names[i].str; i++) {
        if(d_names[i].id == z) {
            return d_names[i].str;
        }
    }
    return "";	
}

char*
OnUpdateClassId(
    ULONG BaseClass
    ) 
{
    for(int i=0; c_names[i].str; i++) {
        if(c_names[i].id == BaseClass) {
            return c_names[i].str;
        }
    }
    return "";	
}

void
PrintBaseAddress(
    PCI_COMMON_CONFIG* pciData,
    ULONG i
    )
{
//    ULONG tmp;

    if(!pciData->u.type0.BaseAddresses[i])
        return;
    printf("\tBase Address %d: %08x\t%s\n", i, pciData->u.type0.BaseAddresses[i] & 0xfffffffe,
        (pciData->u.type0.BaseAddresses[i] & PCI_ADDRESS_IO_SPACE) ? "I/O" : "Mem");

/*    tmp = pciData->u.type0.BaseAddresses[i] & PCI_ADDRESS_MEMORY_TYPE_MASK; \
    SetDlgItemText(IDC_BASEADDRMT##i, \
    (tmp) ? (tmp == 2 ? PCITXT("20 bit") : PCITXT("64 bit")) : PCITXT("32 bit")); \
    pBt = (CButton*)GetDlgItem(IDC_BASEADDR_CHK_PF##i); \
    pBt->SetCheck((CurDevInfo.u.type0.BaseAddresses[i] & PCI_ADDRESS_MEMORY_PREFETCHABLE) ? (h != ((HANDLE)-1)) : 0);*/
}

VOID
PrintDevInfo(
    ULONG              BusId,
    ULONG              DevId,
    ULONG              DevFunc,
    PCI_COMMON_CONFIG* pciData
    )
{
    BOOLEAN mf;
    
    printf("\nBus %02x  Device %02x  Function %02x\n", BusId,  
                                                        DevId,  
                                                        DevFunc);
    printf("    Vendor: %04x\t%s\n", pciData->VendorID, OnUpdateVendorId(pciData->VendorID));
    printf("    Device: %04x\t%s\n", pciData->DeviceID, OnUpdateDeviceId(pciData->VendorID, pciData->DeviceID));
    printf("    Revision: %02x\n", pciData->RevisionID);

    printf("    Device class: %02x\t%s\n", pciData->BaseClass, OnUpdateClassId(pciData->BaseClass));
    printf("    Device subclass: %02x\n", pciData->SubClass);
//    printf("\tDevice sub-subclass: ??\n");

    mf = (pciData->HeaderType & PCI_MULTIFUNCTION) != 0;
    if(mf)
        printf("    Multifunction device\n");

    if((pciData->HeaderType & ~PCI_MULTIFUNCTION) == PCI_DEVICE_TYPE) {

        printf("    Device subvendor: %04x", pciData->u.type0.SubVendorID);
        printf("    Device subsystem: %04x\n", pciData->u.type0.SubSystemID);

        PrintBaseAddress(pciData, 0);
        PrintBaseAddress(pciData, 1);
        PrintBaseAddress(pciData, 2);
        PrintBaseAddress(pciData, 3);
        PrintBaseAddress(pciData, 4);
        PrintBaseAddress(pciData, 5);

        if(pciData->u.type0.ROMBaseAddress) {
            printf("    ROM base: %08x\t%s\n", pciData->u.type0.ROMBaseAddress,
                (pciData->u.type0.ROMBaseAddress & PCI_ROMADDRESS_ENABLED) ? "Enabled" : "Disabled");
        }

        printf("    Interrupt line: %02x\tInterrupt pin: %02x\tMin_Gnt: %02x\tMaxLat: %02x\n",
            pciData->u.type0.InterruptLine,
            pciData->u.type0.InterruptPin,
            pciData->u.type0.MinimumGrant,
            pciData->u.type0.MaximumLatency);
        printf("    Cache line size: %02x\tLatency timer: %02x\tHeader type: %02x\tBIST: %02x\n",
            pciData->CacheLineSize,
            pciData->LatencyTimer,
            pciData->HeaderType,
            pciData->BIST);
        printf("    I/O: %d\tMem: %d\tBusMast: %d\tSpecial: %d\tMemInv: %d\n",
            pciData->Command & PCI_ENABLE_IO_SPACE     ? 1 : 0,
            pciData->Command & PCI_ENABLE_MEMORY_SPACE ? 1 : 0,
            pciData->Command & PCI_ENABLE_BUS_MASTER   ? 1 : 0,
            pciData->Command & PCI_ENABLE_SPECIAL_CYCLES ? 1 : 0,
            pciData->Command & PCI_ENABLE_WRITE_AND_INVALIDATE ? 1 : 0);

    }

}

VOID
BuildDevList()
{
    ULONG BusId=0;
    ULONG DevId=0;
    ULONG DevFunc=0;

    ULONG returned;

    PciDump_GetHeader_USER_IN  buf_in;
    PciDump_GetHeader_USER_OUT buf_out;

    HANDLE h;

    h = CreateFile(NT_PciDump_USER_NAME, GENERIC_READ | GENERIC_WRITE,
                   FILE_SHARE_READ | FILE_SHARE_WRITE,
                   NULL,
                   OPEN_EXISTING,
                   FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);

    while(h != ((HANDLE)-1)) {

        buf_in.BusNumber = BusId;
        buf_in.DeviceNumber = 0;
        buf_in.FunctionNumber = 0;
        
        if(!DeviceIoControl(h,IOCTL_PciDump_GetHeader,
                            &buf_in,sizeof(buf_in),
                            &buf_out,sizeof(buf_out),
                            &returned,NULL)
                  ||
            buf_out.status == 0) {
            CloseHandle(h);
            h = ((HANDLE)-1);
            break;
        }
        
        for(DevId=0; DevId<PCI_MAX_DEVICES; DevId++) {
            
            buf_in.BusNumber = BusId;
            buf_in.DeviceNumber = DevId;
            buf_in.FunctionNumber = 0;

            DeviceIoControl(h,IOCTL_PciDump_GetHeader,
                                &buf_in,sizeof(buf_in),
                                &buf_out,sizeof(buf_out),
                                &returned,NULL);

            if(buf_out.status == 2)
                continue;

/*            buf_in.BusNumber = BusId;
            buf_in.DeviceNumber = DevId;
            buf_in.FunctionNumber = 1;

            DeviceIoControl(h,IOCTL_PciDump_GetHeader,
                                &buf_in,sizeof(buf_in),
                                &buf_out,sizeof(buf_out),
                                &returned,NULL);

            if(buf_out.status == 2)
                continue;*/

            for(DevFunc=0; DevFunc<PCI_MAX_FUNCTION; DevFunc++) {

                buf_in.BusNumber = BusId;
                buf_in.DeviceNumber = DevId;
                buf_in.FunctionNumber = DevFunc;

                DeviceIoControl(h,IOCTL_PciDump_GetHeader,
                                    &buf_in,sizeof(buf_in),
                                    &buf_out,sizeof(buf_out),
                                    &returned,NULL);

                if(buf_out.status == 2)
                    continue;

                // Dump PCI device
                PrintDevInfo(BusId, DevId, DevFunc, &buf_out.pciData);
            }
        }

        BusId++;
    }
      
    if(h != ((HANDLE)-1)) {
        CloseHandle(h);
        h = ((HANDLE)-1);
    }
}

void main (int argc, char* argv[]) {
//    HANDLE h;
    LONG l=0;
    PUCHAR buff=NULL;
    PUCHAR buff_j=NULL;
    PUCHAR buff_i=NULL;
//    LONG tmp;
    LONG i=0, j=0;
    ULONG k;

    if(!Privilege(SE_LOAD_DRIVER_NAME, TRUE)) {
        printf("Warning: Insufficient privileges");
    }
    
    k = NtServiceIsRunning(NT_PciDump_SVC_NAME);
    if(k == 1) {
        // Ok
    } else
    if(k == 0) {
        k = NtServiceStart(NT_PciDump_SVC_NAME);
    }
    load_pci_info();
    if(k >= 0) {
        BuildDevList();
    } else {
        printf("Error: Can't load or install PciPumpEx driver");
        exit(-1);
    }

    exit(0);
}
